

def plot_gc_reads(adata, obs_id, *args, **kwargs):
    """ Plot scatter points of gc by read count.

    Args:
        adata (anndata.AnnData): copy number
        obs_id (str): cell or clone to plot
    """

    raise NotImplementedError
